from __future__ import annotations

import time
from collections.abc import Callable
from typing import Any

from opentelemetry import trace
from opentelemetry.semconv._incubating.attributes.session_attributes import SESSION_ID

from plain.http import Response
from plain.runtime import settings
from plain.utils.cache import patch_vary_headers
from plain.utils.http import http_date

from .core import SessionStore
from .requests import get_request_session, set_request_session


class SessionMiddleware:
    def __init__(self, get_response: Callable[[Any], Any]) -> None:
        self.get_response = get_response

    def __call__(self, request: Any) -> Response:
        session_key = request.cookies.get(settings.SESSION_COOKIE_NAME)

        session = SessionStore(session_key)
        set_request_session(request, session)

        if session.model_instance:
            trace.get_current_span().set_attribute(
                SESSION_ID, session.model_instance.id
            )

        response = self.get_response(request)

        """
        If request.session was modified, or if the configuration is to save the
        session every time, save the changes and set a session cookie or delete
        the session cookie if the session has been emptied.
        """
        session = get_request_session(request)
        accessed = session.accessed
        modified = session.modified
        empty = session.is_empty()

        # First check if we need to delete this cookie.
        # The session should be deleted only if the session is entirely empty.
        if settings.SESSION_COOKIE_NAME in request.cookies and empty:
            response.delete_cookie(
                settings.SESSION_COOKIE_NAME,
                path=settings.SESSION_COOKIE_PATH,
                domain=settings.SESSION_COOKIE_DOMAIN,
                samesite=settings.SESSION_COOKIE_SAMESITE,
            )
            patch_vary_headers(response, ["Cookie"])
        else:
            if accessed:
                patch_vary_headers(response, ["Cookie"])
            if (modified or settings.SESSION_SAVE_EVERY_REQUEST) and not empty:
                if settings.SESSION_EXPIRE_AT_BROWSER_CLOSE:
                    max_age = None
                    expires = None
                else:
                    max_age = settings.SESSION_COOKIE_AGE
                    expires_time = time.time() + max_age
                    expires = http_date(expires_time)
                # Save the session data and refresh the client cookie.
                # Skip session save for 5xx responses.
                if response.status_code < 500:
                    session.save()
                    response.set_cookie(
                        settings.SESSION_COOKIE_NAME,
                        session.session_key,
                        max_age=max_age,
                        expires=expires,
                        domain=settings.SESSION_COOKIE_DOMAIN,
                        path=settings.SESSION_COOKIE_PATH,
                        secure=settings.SESSION_COOKIE_SECURE or None,
                        httponly=settings.SESSION_COOKIE_HTTPONLY or None,
                        samesite=settings.SESSION_COOKIE_SAMESITE,
                    )
        return response



from __future__ import annotations

import string
from collections.abc import Iterator, MutableMapping
from datetime import timedelta
from typing import Any

from plain.models import transaction
from plain.runtime import settings
from plain.utils import timezone
from plain.utils.crypto import get_random_string


class SessionStore(MutableMapping):
    """
    The actual session object that gets attached to a request,
    backed by the underlying Session model for the storage.
    """

    def __init__(self, session_key: str | None = None) -> None:
        self.session_key = session_key
        self.accessed = False
        self.modified = False
        self._session_cache: dict | None = None
        self._session_instance = None

        # Lazy import
        from .models import Session

        self._model = Session

    def __contains__(self, key: object) -> bool:
        return key in self._session

    def __getitem__(self, key: str) -> Any:
        return self._session[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self._session[key] = value
        self.modified = True

    def __delitem__(self, key: str) -> None:
        del self._session[key]
        self.modified = True

    def __iter__(self) -> Iterator[str]:
        return iter(self._session)

    def __len__(self) -> int:
        return len(self._session)

    def clear(self) -> None:
        # To avoid unnecessary persistent storage accesses, we set up the
        # internals directly (loading data wastes time, since we are going to
        # set it to an empty dict anyway).
        self._session_cache = {}
        self._session_instance = None
        self.accessed = True
        self.modified = True

    def is_empty(self) -> bool:
        "Return True when there is no session_key and the session is empty."
        return not self.session_key and not self._session_cache

    def _get_new_session_key(self) -> str:
        "Return session key that isn't being used."
        while True:
            session_key = get_random_string(32, string.ascii_lowercase + string.digits)
            if not self._model.query.filter(session_key=session_key).exists():
                return session_key

    def _get_session_data(self, no_load: bool = False) -> dict:
        """
        Lazily load session from storage (unless "no_load" is True, when only
        an empty dict is stored) and store it in the current instance.
        """
        self.accessed = True

        # If the cache has already been populated (even with an empty dict),
        # simply return it.
        if self._session_cache is not None:
            return self._session_cache

        # The cache hasn't been populated yet so either initialise it to an
        # empty dictionary (when "no_load" is True or there is no session
        # key) or fetch the data from the database.
        if self.session_key is None or no_load:
            self._session_cache = {}
            return self._session_cache

        try:
            session = self._model.query.get(
                session_key=self.session_key, expires_at__gt=timezone.now()
            )
            self._session_instance = session
            self._session_cache = session.session_data
            return self._session_cache
        except self._model.DoesNotExist:
            self.session_key = None
            self._session_instance = None
            self._session_cache = {}
            return self._session_cache

    @property
    def _session(self) -> dict:
        """
        Property to access the session data, ensuring it is loaded.
        """
        return self._get_session_data()

    @property
    def model_instance(self) -> Any:
        """
        Return the underlying Session model instance, or None if no session exists.
        """
        if self._session_instance is not None:
            return self._session_instance

        # Trigger loading of session data which will populate _session_instance
        self._get_session_data()
        return self._session_instance

    def flush(self) -> None:
        """
        Remove the current session data from the database and regenerate the
        key.
        """
        self.clear()
        try:
            self._model.query.get(session_key=self.session_key).delete()
        except self._model.DoesNotExist:
            pass
        self.session_key = None
        self._session_instance = None

    def cycle_key(self) -> None:
        """
        Create a new session key, while retaining the current session data.
        """
        data = self._session
        key = self.session_key
        self.create()
        self._session_cache = data
        if key:
            try:
                self._model.query.get(session_key=key).delete()
            except self._model.DoesNotExist:
                pass

    def create(self) -> None:
        self.session_key = self._get_new_session_key()
        data = self._get_session_data(no_load=True)
        with transaction.atomic():
            self._session_instance = self._model.query.create(
                session_key=self.session_key,
                session_data=data,
                expires_at=timezone.now()
                + timedelta(seconds=settings.SESSION_COOKIE_AGE),
            )
        self.modified = True

    def save(self) -> None:
        """
        Save the current session data to the database using update_or_create.
        """
        data = self._get_session_data(no_load=False)

        with transaction.atomic():
            if self.session_key is None:
                self.session_key = self._get_new_session_key()

            self._session_instance, created = self._model.query.update_or_create(
                session_key=self.session_key,
                defaults={
                    "session_data": data,
                    "expires_at": timezone.now()
                    + timedelta(seconds=settings.SESSION_COOKIE_AGE),
                },
            )

        if created:
            self.modified = True



from __future__ import annotations

from typing import TYPE_CHECKING
from weakref import WeakKeyDictionary

if TYPE_CHECKING:
    from plain.http import Request

    from .core import SessionStore

_request_sessions: WeakKeyDictionary[Request, SessionStore] = WeakKeyDictionary()


def set_request_session(request: Request, session: SessionStore) -> None:
    """Store the session for this request."""
    _request_sessions[request] = session


def get_request_session(request: Request) -> SessionStore:
    """Get the session for this request. Raises KeyError if no session is set."""
    return _request_sessions[request]



from plain.admin.views import (
    AdminModelDetailView,
    AdminModelListView,
    AdminViewset,
    register_viewset,
)

from .models import Session


@register_viewset
class SessionAdmin(AdminViewset):
    class ListView(AdminModelListView):
        model = Session
        fields = ["id", "expires_at", "created_at"]
        search_fields = ["session_key"]
        nav_section = "Sessions"
        nav_icon = "person-badge"
        queryset_order = ["-created_at"]

    class DetailView(AdminModelDetailView):
        model = Session



from plain import models


@models.register_model
class Session(models.Model):
    session_key = models.CharField(max_length=40)
    session_data = models.JSONField(default=dict, required=False)
    created_at = models.DateTimeField(auto_now_add=True)
    expires_at = models.DateTimeField(allow_null=True)

    model_options = models.Options(
        indexes=[
            models.Index(fields=["expires_at"]),
        ],
        constraints=[
            models.UniqueConstraint(fields=["session_key"], name="unique_session_key")
        ],
    )

    def __str__(self) -> str:
        return self.session_key



from plain.preflight import PreflightCheck, PreflightResult, register_check
from plain.runtime import settings


def _session_middleware() -> bool:
    return "plain.sessions.middleware.SessionMiddleware" in settings.MIDDLEWARE


def _session_app() -> bool:
    return "plain.sessions" in settings.INSTALLED_PACKAGES


@register_check(name="sessions.cookie_secure", deploy=True)
class CheckSessionCookieSecure(PreflightCheck):
    """Ensures SESSION_COOKIE_SECURE is True in production deployment."""

    def run(self) -> list[PreflightResult]:
        if settings.SESSION_COOKIE_SECURE is True:
            return []

        warnings = []
        if _session_app():
            warnings.append(
                PreflightResult(
                    fix="You have 'plain.sessions' in your INSTALLED_PACKAGES, but SESSION_COOKIE_SECURE is not set to True. Set SESSION_COOKIE_SECURE=True to prevent session hijacking as using a secure-only session cookie makes it more difficult for network traffic sniffers to hijack user sessions.",
                    id="security.session_cookie_not_secure_app",
                    warning=True,
                )
            )
        if _session_middleware():
            warnings.append(
                PreflightResult(
                    fix="You have 'plain.sessions.middleware.SessionMiddleware' in your MIDDLEWARE, but SESSION_COOKIE_SECURE is not set to True. Set SESSION_COOKIE_SECURE=True to prevent session hijacking as using a secure-only session cookie makes it more difficult for network traffic sniffers to hijack user sessions.",
                    id="security.session_cookie_not_secure_middleware",
                    warning=True,
                )
            )
        if len(warnings) > 1:
            warnings = [
                PreflightResult(
                    fix="SESSION_COOKIE_SECURE is not set to True. Set SESSION_COOKIE_SECURE=True to prevent session hijacking as using a secure-only session cookie makes it more difficult for network traffic sniffers to hijack user sessions.",
                    id="security.session_cookie_not_secure",
                    warning=True,
                )
            ]
        return warnings


@register_check(name="sessions.cookie_httponly", deploy=True)
class CheckSessionCookieHttpOnly(PreflightCheck):
    """Ensures SESSION_COOKIE_HTTPONLY is True in production deployment."""

    def run(self) -> list[PreflightResult]:
        if settings.SESSION_COOKIE_HTTPONLY is True:
            return []

        warnings = []
        if _session_app():
            warnings.append(
                PreflightResult(
                    fix="You have 'plain.sessions' in your INSTALLED_PACKAGES, but SESSION_COOKIE_HTTPONLY is not set to True. Set SESSION_COOKIE_HTTPONLY=True to prevent cross-site scripting attacks as using an HttpOnly session cookie makes it more difficult for cross-site scripting attacks to hijack user sessions.",
                    id="security.session_cookie_not_httponly_app",
                    warning=True,
                )
            )
        if _session_middleware():
            warnings.append(
                PreflightResult(
                    fix="You have 'plain.sessions.middleware.SessionMiddleware' in your MIDDLEWARE, but SESSION_COOKIE_HTTPONLY is not set to True. Set SESSION_COOKIE_HTTPONLY=True to prevent cross-site scripting attacks as using an HttpOnly session cookie makes it more difficult for cross-site scripting attacks to hijack user sessions.",
                    id="security.session_cookie_not_httponly_middleware",
                    warning=True,
                )
            )
        if len(warnings) > 1:
            warnings = [
                PreflightResult(
                    fix="SESSION_COOKIE_HTTPONLY is not set to True. Set SESSION_COOKIE_HTTPONLY=True to prevent cross-site scripting attacks as using an HttpOnly session cookie makes it more difficult for cross-site scripting attacks to hijack user sessions.",
                    id="security.session_cookie_not_httponly",
                    warning=True,
                )
            ]
        return warnings



from plain.sessions.core import SessionStore
from plain.sessions.models import Session
from plain.test import Client


def test_session_created(db):
    assert Session.query.count() == 0

    response = Client().get("/")

    assert response.status_code == 200

    assert Session.query.count() == 1


def test_mapping_attributes(db):
    store = SessionStore()
    assert store.accessed is False
    assert store.modified is False
    assert len(store) == 0

    # accessing without modifying
    assert store.get("foo") is None
    assert store.accessed is True
    assert store.modified is False

    assert "foo" not in store
    assert store.pop("foo", "a") == "a"

    # setting a value modifies the session
    store["foo"] = "bar"
    assert store.modified is True
    assert store["foo"] == "bar"

    # update via MutableMapping.update
    store.update({"a": 1, "b": 2})
    assert set(store.keys()) == {"foo", "a", "b"}
    assert len(store) == 3
    assert store.modified is True

    # deleting a key modifies the session
    del store["foo"]
    assert "foo" not in store
    assert len(store) == 2

    # clearing empties the session
    store.clear()
    assert store.is_empty()
    assert store.modified is True



from __future__ import annotations

from typing import Any

from plain.runtime import settings

from .core import SessionStore


def get_client_session(client: Any) -> SessionStore:
    """Return the current session variables for a test client."""
    cookie = client.cookies.get(settings.SESSION_COOKIE_NAME)
    if cookie:
        return SessionStore(cookie.value)
    session = SessionStore()
    session.save()
    client.cookies[settings.SESSION_COOKIE_NAME] = session.session_key
    return session



from plain.sessions.views import SessionViewMixin
from plain.urls import Router, path
from plain.views import View


class IndexView(SessionViewMixin, View):
    def get(self):
        # Store something so the session is saved
        self.session["foo"] = "bar"
        return "test"


class AppRouter(Router):
    namespace = ""
    urls = [
        path("", IndexView),
    ]



from plain.chores import register_chore
from plain.utils import timezone

from .models import Session


@register_chore("sessions")
def clear_expired() -> str:
    """
    Delete sessions that have expired.
    """
    result = Session.query.filter(expires_at__lt=timezone.now()).delete()
    return f"{result[0]} expired sessions deleted"



# Generated by Plain 0.52.2 on 2025-07-08 01:17

from plain import models
from plain.models import migrations


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Session",
            fields=[
                ("id", models.PrimaryKeyField()),
                ("session_key", models.CharField(max_length=40)),
                ("session_data", models.JSONField(default=dict, required=False)),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("expires_at", models.DateTimeField(allow_null=True)),
            ],
        ),
        migrations.AddIndex(
            model_name="session",
            index=models.Index(
                fields=["expires_at"], name="plainsessio_expires_d87cb5_idx"
            ),
        ),
        migrations.AddConstraint(
            model_name="session",
            constraint=models.UniqueConstraint(
                fields=("session_key",), name="unique_session_key"
            ),
        ),
    ]



from __future__ import annotations

from typing import Any

from jinja2 import pass_context

from plain.templates import register_template_global

from .core import SessionStore
from .requests import get_request_session


@register_template_global
@pass_context
def get_current_session(context: dict[str, Any]) -> SessionStore:
    """Get the session for the current request."""
    request = context.get("request")
    assert request is not None, "No request in template context"
    return get_request_session(request)



from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING

from .requests import get_request_session

if TYPE_CHECKING:
    from plain.http import Request

    from .core import SessionStore


class SessionViewMixin:
    """Mixin that adds session access to views."""

    request: Request

    @cached_property
    def session(self) -> SessionStore:
        """Get the session for this request."""
        return get_request_session(self.request)

    def get_template_context(self) -> dict:
        """Add session to template context."""
        context = super().get_template_context()  # type: ignore
        context["session"] = self.session
        return context
